Package: mockthat
Title: Function Mocking for Unit Testing in R
Version: 0.2.8
Authors@R: 
    person(given = "Nicolas",
           family = "Bennett",
           role = c("aut", "cre"),
           email = "nicolas.bennett@stat.math.ethz.ch")
Description: With the deprecation of mocking capabilities shipped with
    'testthat' as of 'edition 3' it is left to third-party packages to replace
    this functionality, which in some test-scenarios is essential in order to
    run unit tests in limited environments (such as no Internet connection).
    Mocking in this setting means temporarily substituting a function with a
    stub that acts in some sense like the original function (for example by
    serving a HTTP response that has been cached as a file). The only exported
    function 'with_mock()' is modeled after the eponymous 'testthat' function
    with the intention of providing a drop-in replacement.
License: MIT + file LICENSE
URL: https://nbenn.github.io/mockthat/
BugReports: https://github.com/nbenn/mockthat/issues
Depends: R (>= 3.3.0)
Imports: utils, rlang
Suggests: testthat, pkgload, curl, jsonlite, withr
Encoding: UTF-8
Language: en-US
Roxygen: list(markdown = TRUE)
RoxygenNote: 7.2.1
Author: Nicolas Bennett [aut, cre]
Maintainer: Nicolas Bennett <nicolas.bennett@stat.math.ethz.ch>
Built: R 4.4.0; ; 2024-05-01 11:29:05 UTC; unix

Package: mocktest
Title: Package for Testing 'mockthat'
Version: 0.1.0
Authors@R:
    person(given = "Nicolas",
           family = "Bennett",
           role = c("aut", "cre"),
           email = "nicolas.bennett@stat.math.ethz.ch")
Description: As the main purpose of 'mockthat' is mocking functions within
    a package being tested, this dummy package is supplied with 'mockthat'
    for testing purposes.
License: MIT + file LICENSE
Depends:
		curl
Imports:
		utils,
		jsonlite
Encoding: UTF-8
LazyData: true
Roxygen: list(markdown = TRUE)
RoxygenNote: 7.1.1.9000

# Generated by roxygen2: do not edit by hand

S3method(test_s3,default)
S3method(test_s3,test_cls)
export(test_base)
export(test_exported)
export(test_s3)
importFrom(curl,curl_fetch_disk)
importFrom(jsonlite,minify)
importFrom(utils,head)

test_non_exported <- function() {
  test_non_exported_nested()
}

test_non_exported_nested <- function() 10L

test_symbol <- 0L

#' Exported test functions
#'
#' @description
#' These functions only serves to test mocking capabilities of `mockthat`. The
#' exported function `test_exported()` contains a call to the non-exported
#' function `test_non_exported()`, `test_base()` contains calls to base
#' functions [base::isTRUE()] and [base::all.equal()] and `test_s3()` can be
#' used to test mocking in the context of S3 dispatch.
#'
#' @return
#' * `test_exported()`: 10 (as scalar integer)
#' * `test_base()`: Logical flag indicating whether `base::all.equal()`
#'   returned `TRUE`.
#' * `test_s3()`: 10 (as scalar integer) if dispatched on an object of class
#'   `test_cls` and 5 (as scalar integer) if dispatched on any other class
#'
#' @rdname test_funs
#' @export
test_exported <- function() {
  test_non_exported()
}

#' @param x,y Forwarded to [base::all.equal()]
#' @rdname test_funs
#' @export
test_base <- function(x, y) isTRUE(base::all.equal(x, y))

show_struct <- utils::str

#' @importFrom utils head
test_utils <- function(x, how = c("struct", "head", "tail")) {

  switch(
    match.arg(how),
    struct = show_struct(x),
    head = head(x),
    tail = utils::tail(x)
  )
}

#' @rdname test_funs
#' @export
test_s3 <- function(x) UseMethod("test_s3", x)

#' @export
test_s3.test_cls <- function(x) test_non_exported()

#' @export
test_s3.default <- function(x) 5L

curl_dl_mem <- curl::curl_fetch_memory

#' @importFrom curl curl_fetch_disk
test_depends <- function(url, how = c("mem", "disk", "stream")) {

  switch(
    match.arg(how),
    mem = curl_dl_mem(url),
    disk = curl_fetch_disk(url),
    stream = curl::curl_fetch_stream(url)
  )
}

jsonlite_pretty <- jsonlite::prettify

#' @importFrom jsonlite minify
test_imports <- function(x, what = c("pretty", "mini", "conv")) {

  switch(
    match.arg(what),
    pretty = jsonlite_pretty(x),
    mini = minify(x),
    conv = jsonlite::fromJSON(x)
  )
}

pkg_dir <- system.file("testdata", "mocktest", package = "mockthat")
pkg_nme <- pkgload::pkg_name(pkg_dir)

pkgload::load_all(pkg_dir, attach_testthat = FALSE)

withr::defer(pkgload::unload(pkg_nme))
withr::local_envvar(TESTTHAT_PKG = pkg_nme)

test_that("can change value of internal function", {

  val <- test_non_exported()

  res <- mockthat::with_mock(
    test_non_exported_nested = function() 5,
    test_non_exported()
  )

  expect_equal(res, 5)
  expect_error(expect_equal(res, val), class = "expectation_failure")
  expect_equal(test_non_exported(), val)

  # and value is restored on error
  expect_error(
    mockthat::with_mock(
      test_non_exported_nested = function() 5,
      stop("!")
    )
  )

  expect_equal(test_non_exported(), val)
})

test_that("mocks can access local variables", {

  x <- 5

  mockthat::with_mock(
    test_non_exported_nested = function() x,
    expect_equal(test_non_exported(), 5)
  )
})

test_that("non-empty mock with return value", {

  expect_true(
    mockthat::with_mock(test_non_exported = function(...) FALSE, TRUE)
  )
})

test_that("can't mock base functions", {

  expect_error(
    mockthat::with_mock(
      all.equal = function(x, y, ...) TRUE, test_base(3, "a")
    ),
    "Can't mock functions in base package"
  )

  expect_error(
    mockthat::with_mock(
      `base::all.equal` = function(x, y, ...) TRUE, test_base(3, "a")
    ),
    "Can't mock functions in base package"
  )

  expect_error(
    mockthat::with_mock(
      isTRUE = function(x, y, ...) FALSE, test_base(3, "a")
    ),
    "Can't mock functions in base package"
  )

  expect_error(
    mockthat::with_mock(
      `base::isTRUE` = function(x, y, ...) FALSE, test_base(3, "a")
    ),
    "Can't mock functions in base package"
  )
})

test_that("can sometimes mock utils functions", {

  res <- mockthat::with_mock(
    show_struct = "bar",
    test_utils("foo", "struct")
  )

  expect_identical(res, "bar")

  expect_error(
    mockthat::with_mock(str = "bar", test_utils("foo", "struct")),
    "Can't mock functions in utils package"
  )

  expect_error(
    mockthat::with_mock(`utils::str` = "bar", test_utils("foo", "struct")),
    "Can't mock functions in utils package"
  )

  res <- mockthat::with_mock(
    head = "bar",
    test_utils("foo", "head")
  )

  expect_identical(res, "bar")

  expect_error(
    mockthat::with_mock(`utils::head` = "bar", test_utils("foo", "head")),
    "Can't mock functions in utils package"
  )

  expect_error(
    mockthat::with_mock(tail = "bar", test_utils("foo", "tail")),
    "Can't mock functions in utils package"
  )

  expect_error(
    mockthat::with_mock(`utils::tail` = "bar", test_utils("foo", "tail")),
    "Can't mock functions in utils package"
  )
})

test_that("can't mock non-existing", {

  expect_error(
    mockthat::with_mock(..bogus.. = identity, TRUE),
    paste("Function `\\.\\.bogus\\.\\.` not found in environment", pkg_nme)
  )
})

test_that("can't mock non-function", {

  expect_error(
    mockthat::with_mock(test_symbol = FALSE, TRUE),
    paste("Function `test_symbol` not found in environment", pkg_nme)
  )
})

test_that("empty or no-op mock", {

  warn <- paste("Not mocking anything. Please use named arguments to specify",
                "the functions you want to mock.")

  expect_warning(expect_null(mockthat::with_mock()), warn, fixed = TRUE)

  expect_warning(expect_true(mockthat::with_mock(TRUE)), warn, fixed = TRUE)
})

test_that("visibility", {

  expect_warning(
    expect_false(withVisible(mockthat::with_mock())$visible)
  )

  res <- withVisible(
    mockthat::with_mock(
      test_non_exported = function() {},
      TRUE
    )
  )

  expect_true(res$visible)

  res <- withVisible(
    mockthat::with_mock(
      test_non_exported = function() {},
      invisible(5)
    )
  )

  expect_false(res$visible)
})

test_that("multiple return values", {

  expect_true(
    mockthat::with_mock(FALSE, TRUE, test_non_exported = function() {})
  )

  expect_equal(
    mockthat::with_mock(3, test_non_exported = function() {}, 5),
    5
  )
})

test_that("can access variables defined in function", {

  x <- 5

  expect_equal(
    mockthat::with_mock(x, test_non_exported = function() {}),
    5
  )
})

test_that("can mock if package is not loaded", {

  if ("package:jsonlite" %in% search()) {
    skip("jsonlite is loaded")
  }

  skip_if_not_installed("jsonlite")

  mockthat::with_mock(
    `jsonlite::toJSON` = identity,
    expect_identical(jsonlite::toJSON, identity)
  )
})

test_that(
  "changes to variables are preserved between calls and visible outside", {

  x <- 1

  mockthat::with_mock(
    test_non_exported = function() {},
    x <- 3,
    expect_equal(x, 3)
  )

  expect_equal(x, 3)
})

test_that("mock within S3 dispatch", {

  x <- structure(1L, class = "test_cls")

  expect_identical(test_s3(x), 10L)
  expect_identical(mockthat::with_mock(test_non_exported = 5L, test_s3(x)), 5L)

  skip_if(getRversion() < "3.5.0")

  expect_identical(test_s3(1L), 5L)
  expect_identical(mockthat::with_mock(test_s3.default = 2L, test_s3(1L)), 2L)
})

test_that("mock depends", {

  res <- mockthat::with_mock(
    curl_dl_mem = "bar",
    test_depends("foo", "mem")
  )

  expect_identical(res, "bar")

  res <- mockthat::with_mock(
    curl_fetch_disk = "bar",
    test_depends("foo", "disk")
  )

  expect_identical(res, "bar")

  res <- mockthat::with_mock(
    `curl::curl_fetch_stream` = "bar",
    test_depends("foo", "stream")
  )

  expect_identical(res, "bar")
})

test_that("mock imports", {

  res <- mockthat::with_mock(
    jsonlite_pretty = "bar",
    test_imports("foo", "pretty")
  )

  expect_identical(res, "bar")

  res <- mockthat::with_mock(
    minify = "bar",
    test_imports("foo", "mini")
  )

  expect_identical(res, "bar")

  res <- mockthat::with_mock(
    `jsonlite::fromJSON` = "bar",
    test_imports("foo", "conv")
  )

  expect_identical(res, "bar")
})

test_that("mock returns", {

  obj <- "bar"

  res <- mockthat::with_mock(
    curl_dl_mem = obj,
    test_depends("foo", "mem")
  )

  expect_identical(res, obj)

  obj <- c("bar", "baz")

  res <- mockthat::with_mock(
    curl_dl_mem = obj,
    test_depends("foo", "mem")
  )

  expect_identical(res, obj)

  obj <- list(a = 1, b = 2)

  res <- mockthat::with_mock(
    curl_dl_mem = obj,
    test_depends("foo", "mem")
  )

  expect_identical(res, obj)
})

url <- "https://eu.httpbin.org/get?foo=123"

test_that("create and query mock_fun", {

  mk <- mock("mocked request")
  dl <- function(url) curl::curl(url)

  expect_identical(
    with_mock(`curl::curl` = mk, dl(url)),
    "mocked request"
  )

  expect_type(mock_args(mk), "list")
  expect_length(mock_args(mk), 3L)

  expect_identical(mock_arg(mk, "url"), url)
  expect_identical(mock_arg(mk, "open"), "")

  skip_if(getRversion() < "3.6.0")
  expect_identical(mock_call(mk), str2lang("curl::curl(url = url)"))
})

test_that("mock_arg errs if mocked function wasn't called", {
  url <- "https://eu.httpbin.org/get?foo=123"
  mk <- mock("mocked request")
  with_mock(`curl::curl` = mk, {})
  expect_error(
    mock_arg(mk, "url"),
    "mocked function was never called"
  )
})

skip_if_not_installed("withr")

url <- "https://eu.httpbin.org/get?foo=123"

is_ns_env <- function(x) isNamespace(environment(x))

test_that("local mock works", {

  expect_true(is_ns_env(curl::curl))

  dl <- function(url) curl::curl(url)
  mk <- local_mock(`curl::curl` = "mocked request")

  expect_identical(dl(url), "mocked request")
  expect_identical(mock_arg(mk, "url"), url)
  expect_false(is_ns_env(curl::curl))
})

test_that("local mock is restored", {
  expect_true(is_ns_env(curl::curl))
})

test_that("local mock env", {

  dl <- function(url) curl::curl(url)

  local({

    mk <- local_mock(`curl::curl` = "mocked request")

    expect_identical(dl(url), "mocked request")
    expect_identical(mock_arg(mk, "url"), url)
  })
})

url <- "https://eu.httpbin.org/get?foo=123"

test_that("can mock top-level", {

  mok <- function(...) "mocked request"

  res <- with_mock(
    `curl::curl_fetch_memory` = mok,
    curl::curl_fetch_memory(url)
  )

  expect_identical(res, "mocked request")
})

test_that("can access calling env", {

  asw <- "mocked request"
  mok <- function(...) asw

  res <- with_mock(
    `curl::curl_fetch_memory` = mok,
    curl::curl_fetch_memory(url)
  )

  expect_identical(res, asw)
})

test_that("can mock nested", {

  mok <- function(...) "mocked request"
  dlf <- function(x) curl::curl_fetch_memory(x)

  res <- with_mock(
    `curl::curl_fetch_memory` = mok,
    dlf(url)
  )

  expect_identical(res, "mocked request")
})

test_that("visibility propagates", {

  expect_invisible(
    res <- with_mock(
      `curl::curl_fetch_memory` = quote(invisible(NULL)),
      curl::curl_fetch_memory(url)
    )
  )

  expect_null(res)
})

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/funs.R
\name{test_exported}
\alias{test_exported}
\alias{test_base}
\alias{test_s3}
\title{Exported test functions}
\usage{
test_exported()

test_base(x, y)

test_s3(x)
}
\arguments{
\item{x, y}{Forwarded to \code{\link[base:all.equal]{base::all.equal()}}}
}
\value{
\itemize{
\item \code{test_exported()}: 10 (as scalar integer)
\item \code{test_base()}: Logical flag indicating whether \code{base::all.equal()}
returned \code{TRUE}.
\item \code{test_s3()}: 10 (as scalar integer) if dispatched on an object of class
\code{test_cls} and 5 (as scalar integer) if dispatched on any other class
}
}
\description{
These functions only serves to test mocking capabilities of \code{mockthat}. The
exported function \code{test_exported()} contains a call to the non-exported
function \code{test_non_exported()}, \code{test_base()} contains calls to base
functions \code{\link[base:Logic]{base::isTRUE()}} and \code{\link[base:all.equal]{base::all.equal()}} and \code{test_s3()} can be
used to test mocking in the context of S3 dispatch.
}
